

use std::collections::HashSet;
use crate::solver::Solver;
use crate::board::Board;
use std::hash::{Hash, Hasher};
use std::ops::ControlFlow;

static CELLS : usize = 81;

pub struct HechtSolver {
}


#[derive(Hash, Clone, Eq, PartialEq, Debug)]
struct Point {
    pub x: usize,
    pub y: usize,
    pub s: usize,
}

#[derive(Debug)]
struct Action {
    pub p : Point,
    pub value: u16
}

#[derive(Clone, Eq)]
struct SolverBoard {
    arr: [[u16; 9]; 9],
    todos: HashSet<Point>,
    valid: bool,
}

enum SolverBoardItem {
    Recursive(SolverBoard, Vec<Action>),
    Initial(SolverBoard),
}

struct SolverBoardIterator {
    cache: HashSet<[[u16; 9]; 9]>, // to avoid processing the same board twice
    stack: Vec<SolverBoardItem>,
}

impl PartialEq for SolverBoard {
    fn eq(&self, other: &SolverBoard) -> bool {
        self.arr == other.arr && self.valid == other.valid
    }
}

impl Hash for SolverBoard {
    fn hash<H: Hasher>(&self, state: &mut H) {
        self.arr.hash(state);
    }
}

impl HechtSolver {
    pub fn new() -> HechtSolver {
        return HechtSolver {};
    }
}

impl Solver for HechtSolver {
    fn solve(&self, pg: &Board) -> Option<Board> {
        SolverBoardIterator::new(pg)
            .find(|x| x.is_solved())
            .map(|x| x.to_board())
            .flatten()
    }
    
    fn is_unique(&self, pg: &Board) -> bool {
        let first_solution = SolverBoard::new();
        
        SolverBoardIterator::new(pg)
            .filter(|x| x.is_solved())
            .try_fold(first_solution, |prev, x| {
                if !prev.is_solved() || prev == x {
                    ControlFlow::Continue(x)
                } else {
                    ControlFlow::Break(prev)
                }
            }).is_continue()
    }
}

impl SolverBoard {
    
    pub fn new() -> SolverBoard {
        let points = [0usize..CELLS].into_iter().flatten()
            .map(|idx| [idx%9, idx/9])
            .map(|p| Point{x: p[0], y:p[1], s: p[1] / 3 * 3 + p[0] / 3})
            .collect::<HashSet<_>>();
            
        SolverBoard { arr: [[0x1FFu16; 9]; 9], todos: points, valid: true}
    }
    
    pub fn from(pg: &Board) -> SolverBoard {
        let mut spg = SolverBoard::new();

        for p in spg.todos.clone().iter() {
            if let Some(value) = pg.get_value(p.x, p.y) {
                spg.set_value(&p, value);
            }
        }
        
        spg.solve_logically();
        
        spg
    }
    
    fn to_board(&self) -> Option<Board> {
        let mut pg = Board::new();
        
        for x in 0..9 {
            for y in 0..9 {
                let value = self.arr[y][x];
                if value.count_ones() != 1 {
                    return None
                }
                
                pg.set_value(x, y, (value.trailing_zeros() + 1) as u8);
            }
        }
        
        Some(pg)
    }

    fn solve_logically(&mut self) {
        while self.valid {
            if let Some(action) = self.get_simple_action() {
                self.set_bit_value(&action.p, action.value);
            } else {
                break;
            }
        }
    }
    
    fn is_solved(&self) -> bool {
        self.valid && self.todos.len() == 0
    }

    fn set_value(&mut self, p : &Point, value : u8) {
        let bit_value = 1u16 << (value - 1);
        self.set_bit_value(p, bit_value)
    }
    
    fn set_bit_value(&mut self, p : &Point, bit_value : u16) {
        if !self.todos.remove(p) {
            // Point was not in todo-list -> no further action required
            return;
        }

        self.arr[p.y][p.x] = bit_value;
        
        self.todos.iter()
            // Filter positions that do not have the right coordinates
            .filter(|&pos| pos.x == p.x || pos.y == p.y || pos.s == p.s)
            // Filter positions that were known not to hold the value
            .filter(|&pos| (self.arr[pos.y][pos.x] & bit_value) != 0)
            .cloned()
            .collect::<Vec::<_>>()
            .iter()
            .for_each(|pos| self.remove_value(&pos, bit_value));

    }
    
    fn remove_value(&mut self, p : &Point, bit_value: u16) {
        if (self.get_value(p) & bit_value) == 0 {
            return;
        }
        self.arr[p.y][p.x] &= !bit_value;
        
        if let Some(action) = SolverBoard::create_action(p, self.arr[p.y][p.x]) {
            self.set_bit_value(&action.p, action.value);
        } else if self.arr[p.y][p.x] == 0 {
            self.valid = false;
        }
    }
    
    fn create_action(p : &Point, bit_value: u16) -> Option<Action> {
        if bit_value.count_ones() != 1 {
            return None;
        }
        Some(Action{p : p.clone(), value : bit_value})
    }
    
    fn apply(&mut self, action : &Action) {
        self.set_bit_value(&action.p, action.value);
        self.solve_logically();
    }
    
    #[inline]
    fn get_bitvalue_sequence() -> &'static[u16; 9] {
        return &[0b0_0000_0001u16, // 1
                 0b0_0000_0010u16, // 2
                 0b0_0000_0100u16, // 3
                 0b0_0000_1000u16, // 4
                 0b0_0001_0000u16, // 5
                 0b0_0010_0000u16, // 6
                 0b0_0100_0000u16, // 7
                 0b0_1000_0000u16, // 8
                 0b1_0000_0000u16]; // 9
    }
    
    #[inline]
    fn get_value(&self, p : &Point) -> u16 {
        return self.arr[p.y][p.x];
    }

    fn get_simple_action(&self) -> Option<Action> {
        
        // println!("Simple Action");
        self.todos.iter()
            .find_map(|lhs| {
                    let own_value = self.get_value(lhs);
                    self.todos.iter()
                        .filter(|&rhs| lhs.x == rhs.x || lhs.y == rhs.y || lhs.s == rhs.s)
                        .filter(|&rhs| lhs != rhs)
                        .map(|rhs| {
                            let value = self.get_value(&rhs) & own_value;
                            let row_value = if lhs.x == rhs.x {value} else {0};
                            let col_value = if lhs.y == rhs.y {value} else {0};
                            let sec_value = if lhs.s == rhs.s {value} else {0};
                            
                            [row_value, col_value, sec_value]
                        } )
                        .reduce(|a,b| [a[0]|b[0], a[1]|b[1], a[2]|b[2]])
                        .iter()
                        .flatten()
                        .map(|value| value ^ own_value)
                        .filter(|value| value.count_ones() == 1)
                        .find_map(|value| SolverBoard::create_action(lhs, value))
                } )
    }

    fn get_all_actions(&self) -> Vec<Action> {
        if self.valid && !self.is_solved() {
            let mut vec = self.todos.iter()
                .map(|point| Action{p: point.clone(), value: self.get_value(point)})
                .collect::<Vec<_>>(); // collect into the vector so we can (optionally) apply sorting

            // now sort it (lowest value should be at the end (we use it as a stack))
            vec.sort_unstable_by_key(|action| CELLS as u32 - action.value.count_ones());
            
            return vec.iter()
                        .map(|action| SolverBoard::get_bitvalue_sequence().iter()
                                     .filter(|&value| action.value & value != 0)
                                     .map(|&value| Action{p: action.p.clone(), value})
                                     .collect::<Vec<_>>()
                        )
                        .flatten()
                        .collect::<Vec<_>>();
        }
        Vec::new()
    }

}

impl SolverBoardIterator {
    fn new(board : &Board) -> SolverBoardIterator {
        SolverBoardIterator::from_board(SolverBoard::from(board))
    }
    
    fn from_board(board: SolverBoard) -> SolverBoardIterator {
        SolverBoardIterator{cache: HashSet::with_capacity(2048), stack: vec![SolverBoardItem::Initial(board)]}
    }
    
    fn next_board(&mut self) -> Option<SolverBoard> {
        let mut result : Option<SolverBoard> = None;
        while let Some(mut item) = self.stack.pop() {
            if let Some(action) = item.pop_action() {
                if let Some(board) = item.get_board() {
                    let mut new_board = board.clone();
                    self.stack.push(item);
                    new_board.apply(&action);
                    result = self.add_recursive(new_board);
                }
            } else if let Some(board) = item.get_board() { // initial
                result = self.add_recursive(board.clone());
            }

            if result.is_some() {
                return result
            }
        }
        
        println!("Cache size: {}", self.cache.len());
        None
    }
    
    fn add_to_cache(&mut self, board: &SolverBoard) -> Option<SolverBoard> {
        if !self.cache.contains(&board.arr) {
            self.cache.insert(board.arr);
            return Some(board.clone())
        }
        None
    }
    
    fn add_recursive(&mut self, board: SolverBoard) -> Option<SolverBoard> {
        if board.valid {
            let result = self.add_to_cache(&board);
            if result.is_some() {
                let actions = board.get_all_actions();
                self.stack.push(SolverBoardItem::Recursive(board, actions));
            }
            return result;
        }
        None
    }
    
}

impl SolverBoardItem {
    
    fn get_board(&self) -> Option<&SolverBoard> {
        if let SolverBoardItem::Recursive(board, _) = self {
            return Some(board)
        } else if let SolverBoardItem::Initial(board) = self {
            return Some(board)
        }
        None
    }
    
    fn pop_action(&mut self) -> Option<Action> {
        if let SolverBoardItem::Recursive(_, actions) = self {
            return actions.pop()
        }
        None
    }
}

impl Iterator for SolverBoardIterator {
    type Item = SolverBoard;
    
    fn next(&mut self) -> Option<Self::Item> {
        self.next_board()
    }
        
}